@@ -1,4 +1,4 @@
-/* $Id: client.c,v 0.5 1996/09/22 03:35:56 yoichi v0_33 yoichi $
+/* $Id: client.c,v 0.10 1997/03/17 09:36:17 yoichi 0_6 $
  *
  * dhcpcd - DHCP client daemon -
  * Copyright (C) 1996 Yoichi Hariguchi
@@ -46,6 +46,8 @@
 #include "hostinfo.h"
 #include "client.h"
 
+extern int BeRFC1541;			/* if 1, Be RFC154 compliant */
+
 
 int (*Fsm[MAX_STATES])();		/* finite state machine */
 int CurrState;					/* current state */
@@ -178,12 +180,14 @@ initReboot()
 int
 init()
 {
+	char ifname[IFNAMSIZ];
 	int waitTime = 0;
 	struct sockaddr_in	addr;
 
 	freeOptInfo(OptOffer);		/* clear up 'OptOffer' */
 	if ( PrevState == RENEWING || PrevState == REBINDING ) {
-		ifReset(Ifbuf.ifname);
+		strcpy(ifname, Ifbuf.ifname);
+		ifReset(ifname);
 		dhcpMsgInit(Ifbuf.ifname);
 	}
 	mkDhcpDiscoverMsg(Ifbuf.haddr, &DhcpMsgSend);
@@ -236,6 +240,14 @@ rebooting()
 		}
 		/* now the client is initialized
 		 */
+
+		/* initHost uses the values of Ifbuf.bcast and Ifbuf.mask as is
+		 * under the following condition:
+		 *   1. the DHCP ACK msg does not includes bcast or subnetmask option
+		 *   2. Ifbuf.bcast or Ifbuf.mask is not 0.
+		 * So we have to set these values to 0 not to confuse initHost.
+		 */
+		Ifbuf.bcast = Ifbuf.mask = 0;
 		initHost(&Ifbuf, DhcpMsgRecv.yiaddr);
 		execCommandFile();
 		if ( LeaseTime == INFINITE_LEASE_TIME ) {
@@ -270,7 +282,8 @@ selecting()
 		ServerInaddr = *((u_int *)(OptPtr[OserverInaddr]+1));
 		LeaseTime	 = *((u_int *)(OptPtr[OleaseTime]+1));
 		mkDhcpRequestMsg(SELECTING, ServerInaddr, LeaseTime,
-						 DhcpMsgRecv.xid, DhcpMsgRecv.yiaddr, &DhcpMsgSend);
+							 DhcpMsgRecv.xid, DhcpMsgRecv.yiaddr,
+							 &DhcpMsgSend);
 #ifdef NEED_BCAST_RESPONSE
 		DhcpMsgSend.flags = htons(F_BROADCAST);
 #endif
@@ -382,17 +395,19 @@ renewing()
 		time(&sendTime);
 		if ( sendto(Ssend, (char *)&DhcpMsgSend, sizeof(DhcpMsgSend), 0,
 					(struct sockaddr *)&addr, sizeof(addr)) < 0 ) {
-			logSysExit("sendto (renewing)");
-		}
-		/* wait server's response
-		*/
-		if ( rcvAndCheckDhcpMsg(Srecv, &DhcpMsgRecv,
-								 waitMsgType, OptPtr, tm) ) {
-			nextState = setDhcpInfo(OptPtr, &DhcpMsgRecv);
-			if ( nextState == EXCEPTION ) {
-				return nextState;
+			logSysRet("sendto (renewing)");
+			sleep(tm/1000000);
+		} else {
+			/* wait server's response
+			 */
+			if ( rcvAndCheckDhcpMsg(Srecv, &DhcpMsgRecv,
+									waitMsgType, OptPtr, tm) ) {
+				nextState = setDhcpInfo(OptPtr, &DhcpMsgRecv);
+				if ( nextState == EXCEPTION ) {
+					return nextState;
+				}
+				break;
 			}
-			break;
 		}
 		tm = getNextTimeout(NEXT_TIMEOUT);	/* renew response timeout value */
 		timeout -= time(NULL) - prevTime;
@@ -422,6 +437,8 @@ rebinding()
 
 	logRet("got in REBINDING state");
 
+	nextState = INIT;			/* init nextState */
+
 	/* setup for sending broadcast DHCPREQUEST message
 	 */
 	setSockAddrIn(htons(DHCP_SERVER_PORT), Ifbuf.bcast, &addr);
@@ -442,15 +459,17 @@ rebinding()
 		time(&sendTime);
 		if ( sendto(Ssend, (char *)&DhcpMsgSend, sizeof(DhcpMsgSend), 0,
 					(struct sockaddr *)&addr, sizeof(addr)) < 0 ) {
-			logSysExit("sendto (rebinding)");
-		}
-		if ( rcvAndCheckDhcpMsg(Srecv, &DhcpMsgRecv,
-								 waitMsgType, OptPtr, tm) ) {
-			nextState = setDhcpInfo(OptPtr, &DhcpMsgRecv);
-			if ( nextState == EXCEPTION ) {
-				return nextState;
+			logSysRet("sendto (rebinding)");
+			sleep(tm/1000000);
+		} else {
+			if ( rcvAndCheckDhcpMsg(Srecv, &DhcpMsgRecv,
+									waitMsgType, OptPtr, tm) ) {
+				nextState = setDhcpInfo(OptPtr, &DhcpMsgRecv);
+				if ( nextState == EXCEPTION ) {
+					return nextState;
+				}
+				break;
 			}
-			break;
 		}
 		tm = getNextTimeout(NEXT_TIMEOUT);
 		timeout -= time(NULL) - prevTime;
@@ -501,10 +520,11 @@ mkDhcpDiscoverMsg(u_char *haddr, dhcpMessage *msg)
 		p += sizeof(long);
 	}
 	*p++ = dhcpParamRequest;	/* Parameter Request List */
-	*p++ = 7;					/* number of requests */
+	*p++ = 8;					/* number of requests */
 	*p++ = subnetMask;
 	*p++ = routersOnSubnet;
 	*p++ = dns;
+	*p++ = hostName;
 	*p++ = domainName;
 	*p++ = broadcastAddr;
 	*p++ = ntpServers;
@@ -528,12 +548,21 @@ mkDhcpRequestMsg(int flag, u_long serverInaddr, u_long leaseTime,
 	msg->flags = htons(0);		/* do not set the broadcast flag here */
 	bzero((char *)p, sizeof(msg->options) - 4);	/* clear DHCP option field */
 
-	if ( flag == REBOOTING || flag == SELECTING ) {
-		msg->ciaddr = htonl(0);	/* ciaddr must be 0 in REBOOTING & SELECTING */
-		*p++ = dhcpRequestedIPaddr;
-		*p++ = 4;
-		*((u_int *)p) = ciaddr;
-		p += sizeof(u_int);
+	/* 1. Requested IP address must not be in the DHCPREQUEST message
+	 *    under the RFC1541 mode.
+	 * 2. Requested IP address must be in the DHCPREQUEST message
+	 *    sent in the SELECTING or INIT-REBOOT state under the Internet
+	 *    Draft mode
+	 */
+	if ( !BeRFC1541 ) {
+		if ( flag == REBOOTING || flag == SELECTING ) {
+			/* ciaddr must be 0 in REBOOTING & SELECTING */
+			msg->ciaddr = htonl(0);
+			*p++ = dhcpRequestedIPaddr;
+			*p++ = 4;
+			*((u_int *)p) = ciaddr;
+			p += sizeof(u_int);
+		}
 	}
 	*p++ = dhcpMessageType;				/* DHCP message type */
 	*p++ = 1;
@@ -542,7 +571,13 @@ mkDhcpRequestMsg(int flag, u_long serverInaddr, u_long leaseTime,
 	*p++ = 2;
 	*((u_short *)p) = htons(sizeof(*msg));
 	p += sizeof(u_short);
-	if ( flag == REQUESTING) {
+	if ( flag == REBOOTING && SuggestLeaseTime ) {
+		*p++ = dhcpIPaddrLeaseTime;	/* IP address lease time */
+		*p++ = 4;
+		*((u_int *)p) = htonl(SuggestLeaseTime);
+		p += sizeof(long);
+	}
+	if ( flag == SELECTING || flag == REQUESTING ) {
 		*p++ = dhcpServerIdentifier;	/* server identifier */
 		*p++ = 4;
 		*((u_int *)p) = serverInaddr;
@@ -554,6 +589,17 @@ mkDhcpRequestMsg(int flag, u_long serverInaddr, u_long leaseTime,
 		*((u_int *)p) = leaseTime;
 		p += sizeof(u_int);
 	}
+	*p++ = dhcpParamRequest;	/* Parameter Request List */
+	*p++ = 8;					/* number of requests */
+	*p++ = subnetMask;
+	*p++ = routersOnSubnet;
+	*p++ = dns;
+	*p++ = hostName;
+	*p++ = domainName;
+	*p++ = broadcastAddr;
+	*p++ = ntpServers;
+	*p++ = nisDomainName;
+
 	*p++ = dhcpClassIdentifier;			/* class identifier */
 	*p++ = strlen(ClassId);
 	strcpy(p, ClassId);
@@ -568,7 +614,15 @@ mkDhcpDeclineMsg(int flag, u_long serverInaddr, u_long ciaddr,
 	u_char *p =	msg->options + 4; /* just after the magic cookie */
 
 	msg->xid	= rand();
-	msg->ciaddr = ciaddr;
+	if ( flag == DHCP_RELEASE ) {	/* make a DHCPRELEASE msg */
+		msg->ciaddr = ciaddr;
+	} else {						/* make a DHCPDECLINE msg */
+		if ( BeRFC1541 ) {			/* use ciaddr in RFC1541 compliant mode */
+			msg->ciaddr = ciaddr;
+		} else {
+			msg->ciaddr = 0;
+		}
+	}
 
 	bzero((char *)p, sizeof(msg->options) - 4);
 	*p++ = dhcpMessageType;		/* DHCP message type */
@@ -578,6 +632,15 @@ mkDhcpDeclineMsg(int flag, u_long serverInaddr, u_long ciaddr,
 	*p++ = 4;
 	*((u_int *)p) = serverInaddr;
 	p += sizeof(long);
+	if ( flag == DHCP_DECLINE && !BeRFC1541 ) {
+		/* use the requested IP address option
+		 * in the Internet Draft compliant mode
+		 */
+		*p++ = dhcpRequestedIPaddr;
+		*p++ = 4;
+		*((u_int *)p) = ciaddr;
+		p += sizeof(u_int);
+	}
 	*p = endOption;				/* end */
 }
 
@@ -617,12 +680,12 @@ setDhcpInfo(u_char *optp[], dhcpMessage *msg)
 		ServerInaddr = *((u_long *)(optp[OserverInaddr]+1));
 		LeaseTime	 = *((u_long *)(optp[OleaseTime]+1));
 		if ( optp[OrenewalTime] == NULL ) {
-			RenewTime = LeaseTime / 2;		/* default T1 time */
+			RenewTime = htonl(ntohl(LeaseTime) / 2);		/* default T1 time */
 		} else {
 			RenewTime = *((u_long *)(optp[OrenewalTime]+1));
 		}
 		if ( optp[OrebindTime] == NULL ) {
-			RebindTime = LeaseTime / 8 * 7;	/* default T2 time */
+			RebindTime = htonl(ntohl(LeaseTime) / 8 * 7);	/* default T2 time */
 		} else {
 			RebindTime = *((u_long *)(optp[OrebindTime]+1));
 		}
@@ -694,11 +757,12 @@ initHost(struct ifinfo *ifbuf, u_long yiaddr)
 		logRet("assigned broadcast address: %s",
 			   inet_ntoa(*((struct in_addr *)&ifbuf->bcast)));
 	}
-	ifbuf->flags = IFF_UP | IFF_BROADCAST | IFF_NOTRAILERS | IFF_RUNNING;
+	ifbuf->flags =
+		IFF_UP | IFF_BROADCAST | IFF_NOTRAILERS | IFF_RUNNING | IFF_MULTICAST;
 	ifConfig(ifbuf);
 
 	saveIfInfo(ifbuf);			/* save interface information onto file */
-	saveHostInfo((const u_char **)OptOffer);
+	saveHostInfo((const u_char **)OptPtr);
 }
 
 long
